use rand::{distributions::Uniform, Rng};

use crate::{
    ast::{Expr, Ops, TopLevel},
    error::RunError,
};

/// Module level result for returning `T` or a [`RunError`]
/// 
/// [`RunError`]: crate::error::RunError
pub type Result<T> = std::result::Result<T, RunError>;

#[derive(Debug)]
/// Final output of completely evaluated shedron
pub struct Output {
    pub rolls: Vec<i128>,
}

#[derive(Debug)]
/// All possible output values of an evaluated expression
pub enum OutputValues {
    Number(i128),
    Rolls(Vec<i128>),
}

impl OutputValues {
    /// Returns `true` if the output values is [`Rolls`].
    ///
    /// [`Rolls`]: OutputValues::Rolls
    #[must_use]
    pub fn is_rolls(&self) -> bool {
        matches!(self, Self::Rolls(..))
    }
}


/// Simple unit struct to contain interpreter function definitions
pub struct Interpreter;

impl Interpreter {
    pub fn new() -> Self {
        Self
    }

    /// Runs shedron from an AST
    pub fn run(&self, ast: TopLevel) -> Result<Output> {
        let expr = ast.expr;
        let out = self.eval_expr(expr)?;

        let rolls = match out {
            OutputValues::Number(n) => vec![n],
            OutputValues::Rolls(r) => r,
        };

        Ok(Output { rolls })
    }

    /// Evaluates a single expression to [`Result<OutputValues>`]
    /// 
    /// [`OutputValues`]: Result<crate::interpret::OutputValues>
    pub fn eval_expr(&self, expr: Expr) -> Result<OutputValues> {
        match expr {
            crate::ast::Expr::Op(op) => {
                let op = op.as_ref();

                self.eval_op(op)
            }
            crate::ast::Expr::Parens(e) => {
                self.eval_expr(*e)
            },
            crate::ast::Expr::Sum(e) => {
                let expr = self.eval_expr(*e)?;
                match &expr {
                    OutputValues::Rolls(r) => Ok(OutputValues::Number(r.into_iter().sum())),
                    OutputValues::Number(..) => Ok(expr)
                }
            }
            crate::ast::Expr::Number(n) => Ok(OutputValues::Number(n)),
            crate::ast::Expr::Empty => unreachable!("oh no"),
        }
    }

    /// Evaluates a single operator to [`Result<OutputValues>`]
    /// 
    /// [`OutputValues`]: Result<crate::interpret::OutputValues>
    pub fn eval_op(&self, op: &Ops) -> Result<OutputValues> {
        match op {
            Ops::Add(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l + r))
                    }

                    #[cfg(all(feature = "ravy", not(feature = "krista-nikky")))]
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.push(r);
                        Ok(OutputValues::Rolls(l))
                    }

                    #[cfg(all(feature = "krista-nikky", not(feature = "ravy")))]
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n + r).collect();
                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Sub(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l - r))
                    }

                    #[cfg(all(feature = "ravy", not(feature = "krista-nikky")))]
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.push(-r);
                        Ok(OutputValues::Rolls(l))
                    }

                    #[cfg(all(feature = "krista-nikky", not(feature = "ravy")))]
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n - r).collect();
                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Mul(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l * r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n * r).collect();
                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::RDiv(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l / r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n / r).collect();
                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::FDiv(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l / r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n / r).collect();
                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Max(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(if l > r { l } else { r }))
                    }
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.sort_by(|a, b| b.cmp(a));
                        let l = l.into_iter().take(r as usize).collect::<Vec<_>>();

                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Min(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(if l < r { l } else { r }))
                    }
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.sort_by(|a, b| a.cmp(b));
                        let l = l.into_iter().take(r as usize).collect::<Vec<_>>();

                        Ok(OutputValues::Rolls(l))
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Roll(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        if l <= 0 {
                            return Err(RunError::new("Left hand side of roll is less than or equal to 0."))
                        } else if r <= 0 {
                            return Err(RunError::new("Right hand side of roll is less than or equal to 0."))
                        } 
                        if r == 1 {
                            let vals: Vec<i128> = (0..l).map(|_| 1).collect();
                            Ok(OutputValues::Rolls(vals))
                        } else {
                            let mut rng = rand::thread_rng();
                            let range = Uniform::new_inclusive(1, r);

                            let vals: Vec<i128> = (0..l).map(|_| rng.sample(&range)).collect();

                            Ok(OutputValues::Rolls(vals))
                        }
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    fn eval_dice(src: &str) {
        let mut lexer = crate::parser::Lexer::new(src);
        let interpreter = Interpreter::new();

        let _ = dbg!(interpreter.run(lexer.parse().unwrap()));
    }

    #[test]
    fn test_run() {
        eval_dice("2d1");
        eval_dice("2d[1d2]");
        eval_dice("(10d6)+5");
        eval_dice("[1d6]d[1d6]");
        eval_dice("([6d([1d6])]_96*[1d6])//7");
    }
}
